{

    Game.pp                            last modified: 5 July 2014

    Copyright (C) Jaroslaw Baran, furious programming 2011 - 2014.
    All rights reserved.
   __________________________________________________________________________

    This unit is a part of the SnakeASCII console game.
   __________________________________________________________________________

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
   __________________________________________________________________________

}

unit Game;

{$MODE OBJFPC}{$H+}

interface

uses
  Crt, SysUtils, Painter, Converter, Controller, Timer, Level, Snake, Gold, Common;


type
  { game class }
  TGame = class(TObject)
  private
    FAppPath: AnsiString;
    FSnake: TSnake;
    FLevel: TLevel;
    FGold: TGold;
    FPainter: TPainter;
    FController: TController;
    FTimer: TTimer;
  private
    procedure Reset();
    procedure Pause();
    procedure HandleSnakeMove(ADirection: TMoveDirection);
  public
    constructor Create();
    destructor Destroy(); override;
  public
    procedure Play();
  end;


{ ----- end interface ----------------------------------------------------------------------------------------- }


implementation


constructor TGame.Create();
begin
  inherited Create();

  { set the application path }
  FAppPath := ExtractFilePath(ParamStr(0));

  { create level }
  FLevel := TLevel.Create(FAppPath + 'levels\0.lvl');
  { create snake }
  FSnake := TSnake.Create(FLevel.SnakeDefHeadPos, FLevel.SnakeDefDir);
  { create gold }
  FGold := TGold.Create();

  { create game painter }
  FPainter := TPainter.Create();
  { pass the objects reference to painter }
  FPainter.PassSnake(FSnake);
  FPainter.PassLevel(FLevel);
  FPainter.PassGold(FGold);

  { create game controller }
  FController := TController.Create();

  { create game timer }
  FTimer := TTimer.Create();
  FTimer.MaxStepsCount := FLevel.SnakeSpeed;
end;


destructor TGame.Destroy();
begin
  { destroy snake }
  FSnake.Free();
  { destroy level }
  FLevel.Free();
  { destroy gold }
  FGold.Free();

  { destroy game painter }
  FPainter.Free();
  { destroy game controller }
  FController.Free();
  { destroy game timer }
  FTimer.Free();

  inherited Destroy();
end;


procedure TGame.Reset();
begin
  ClrScr();

  { initialize the snake }
  FSnake.Reset(FLevel.SnakeDefHeadPos, FLevel.SnakeDefDir);
  { set the gold initial coordinates }
  FGold.RandomNewCoords(@FLevel.IsInLevel, @FSnake.IsNotOnSnake);

  { draw level bounds and grass clumps }
  FPainter.DrawLevel();
  { draw snake }
  FPainter.DrawLiveSnake();

  { draw gold }
  FPainter.DrawGold();
  { set the gold drawn state }
  FGold.Drawn := True;

  { initialize the game controller }
  FController.ResetInput();

  { initialize the game timer cycles counter }
  FTimer.Reset();
  { sleep the game for a while }
  Delay(START_GAME_DELAY_TIME);
end;


procedure TGame.Pause();
begin
  { draw the PAUSE text }
  FPainter.DrawText(38, 4, White, 'PAUSE');

  { wait for Reset input key }
  repeat
    { reset the input }
    FController.ResetInput();

    { wait for input key }
    while not FController.InputKeyPressed do
    begin
      { sleep the frame }
      Delay(SINGLE_FRAME_DELAY_TIME);
      { process the input }
      FController.ProcessInput();
    end;
  until FController.InputKey = INPUT_DEFAULT_KEY_PAUSE;

  { reset the input }
  FController.ResetInput();

  { clear the PAUSE text }
  FPainter.ClearText(38, 4, 5);
end;


procedure TGame.HandleSnakeMove(ADirection: TMoveDirection);
var
  mcNew: TModuleCoords;
  boolCaughtGold: Boolean;
begin
  { translate move direction to new module coordinates }
  mcNew := TConverter.MoveDirectionToModuleCoords(ADirection, FSnake.HeadCoords);

  { valid move }
  if FSnake.CanMove(mcNew) and FLevel.CanMove(mcNew) then
  begin
    { snake caught the gold }
    boolCaughtGold := FGold.IsGoldModule(mcNew);

    { if snake does not grab the gold }
    if not boolCaughtGold then
      { clear snake tail }
      FPainter.ClearSnakeTail();

    { move snake }
    FSnake.Move(ADirection, boolCaughtGold);
    { draw snake in new coordinates }
    FPainter.DrawLiveSnake();

    { check the caught gold state }
    if boolCaughtGold then
    begin
      { random new gold coordinates }
      FGold.RandomNewCoords(@FLevel.IsInLevel, @FSnake.IsNotOnSnake);
      { draw gold in new coordinates }
      FPainter.DrawGold();
    end;
  end
  else
    { colision }
    begin
      { draw dead snake }
      FPainter.DrawDeadSnake();

      { draw the RESET text }
      FPainter.DrawText(33, 4, White, 'Press R to reset');

      { wait for Reset input key }
      repeat
        { reset the input }
        FController.ResetInput();

        { wait for input key }
        while not FController.InputKeyPressed do
        begin
          { sleep the frame }
          Delay(SINGLE_FRAME_DELAY_TIME);
          { process the input }
          FController.ProcessInput();
        end;
      until FController.InputKey = INPUT_DEFAULT_KEY_RESET;

      { reset the game }
      Self.Reset();
    end;
end;


procedure TGame.Play();
var
  ikInput: TInputKey;
  mdInput: TMoveDirection;
begin
  { initialize the game }
  Self.Reset();

  { main game loop }
  while SNAKE_IS_ALIVE do
  begin
    { step the timer, but reset it first if necessary }
    FTimer.Step();

    { process the input if key is not pressed }
    if not FController.InputKeyPressed then
      FController.ProcessInput();

    { get the input key }
    ikInput := FController.InputKey;
    { translate input key to move direction }
    mdInput := TConverter.InputKeyToMoveDirection(ikInput);

    { detect permanent turbo key }
    if ikInput = INPUT_DEFAULT_KEY_TURBO then
      { invert the permanent turbo state }
      FSnake.InvertPermanentTurbo();

    { opposite or undefined direction }
    if (mdInput = mdNone) or TConverter.OppositeDirection(FSnake.Direction, mdInput) then
      { use the same direction as current }
      mdInput := FSnake.Direction
    else
      { same direction }
      if TConverter.SameDirection(FSnake.Direction, mdInput) then
        { turn on the turbo mode }
        ikInput := INPUT_DEFAULT_KEY_TURBO;

    { handle the input key }
    case ikInput of
      { pause the game }
      INPUT_DEFAULT_KEY_PAUSE: Self.Pause();
      { reset the game }
      INPUT_DEFAULT_KEY_RESET: Self.Reset();
      { quit the game }
      INPUT_DEFAULT_KEY_QUIT: Exit();
    else
      { same direction key or turbo key was pressed, eventually if it's time for handling the snake move }
      if (ikInput = INPUT_DEFAULT_KEY_TURBO) or FSnake.PermanentTurbo or FTimer.ItsTime() then
      begin
        { move the snake }
        HandleSnakeMove(mdInput);
        { reset the input }
        FController.ResetInput();
      end;
    end;
  end;
end;


{ ----- end implementation ------------------------------------------------------------------------------------ }


end.

