{

    Level.pp                           last modified: 2 July 2014

    Copyright (C) Jaroslaw Baran, furious programming 2011 - 2014.
    All rights reserved.
   __________________________________________________________________________

    This unit is a part of the SnakeASCII console game.
   __________________________________________________________________________

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
   __________________________________________________________________________

}

unit Level;

{$MODE OBJFPC}{$H+}{$PACKENUM 1}

interface

uses
  SysUtils, IniFiles, Common;


type
  { level main information redord }
  TLevelInfoRec = record
    Name: AnsiString;
    Difficulty: TLevelOfDifficulty;
  end;

type
  { snake information record }
  TSnakeInfoRec = record
    Coloration: TSnakeColoration;
    Speed: TLevelOfSnakeSpeed;
    DefHeadCoords: TModuleCoords;
    DefDirection: TMoveDirection;
  end;

type
  { map information record }
  TMapInfoRec = record
    BoundsColor: UInt8;
    Map: TLevelMapArr;
  end;

type
  { holes information record }
  THolesInfoRec = record
    Count: Integer;
    Holes: TLevelHolesArr;
  end;

type
  { grass clump information record }
  TGrassClumpsInfoRec = record
    Count: Integer;
    Clumps: TLevelGrassClumpsArr;
  end;

type
  { single level class }
  TLevel = class(TObject)
  private
    FLevelInfo: TLevelInfoRec;
    FSnakeInfo: TSnakeInfoRec;
    FMapInfo:   TMapInfoRec;
    FHolesInfo: THolesInfoRec;
    FGrassInfo: TGrassClumpsInfoRec;
  private
    function LetterToLevelDiff(ALetter: AnsiChar): TLevelOfDifficulty;
    function LetterToSnakeColoration(ALetter: AnsiChar): TSnakeColoration;
    function LetterToSnakeSpeed(ALetter: AnsiChar): TLevelOfSnakeSpeed;
    function LetterToSnakeMoveDir(ALetter: AnsiChar): TMoveDirection;
    function LetterToMapChar(ALetter: AnsiChar): AnsiChar;
  private
    function GetSnakeSpeed(): UInt8;
    function GetMapModuleChar(AX, AY: UInt8): AnsiChar;
    function GetGrassClump(AIndex: UInt8): TLevelGrassClumpRec;
    function CoordsInHole(ACoords: TModuleCoords): Boolean;
  private
    procedure LoadFromFile(AFileName: AnsiString);
  public
    constructor Create(ALevelFileName: AnsiString);
    destructor Destroy(); override;
  public
    function CanMove(ACoords: TModuleCoords): Boolean;
    function IsInLevel(ACoords: TModuleCoords): Boolean;
  public
    property Name: AnsiString read FLevelInfo.Name;
    property Difficulty: TLevelOfDifficulty read FLevelInfo.Difficulty;
    property SnakeColoration: TSnakeColoration read FSnakeInfo.Coloration;
    property SnakeSpeed: UInt8 read GetSnakeSpeed;
    property SnakeDefHeadPos: TModuleCoords read FSnakeInfo.DefHeadCoords;
    property SnakeDefDir: TMoveDirection read FSnakeInfo.DefDirection;
    property BoundsColor: UInt8 read FMapInfo.BoundsColor;
    property ModuleChar[AX, AY: UInt8]: AnsiChar read GetMapModuleChar;
    property ClumpsCount: Integer read FGrassInfo.Count;
    property Clump[AIndex: UInt8]: TLevelGrassClumpRec read GetGrassClump;
  end;


{ ----- end interface ----------------------------------------------------------------------------------------- }


implementation


constructor TLevel.Create(ALevelFileName: AnsiString);
begin
  inherited Create();

  { load level map from file }
  LoadFromFile(ALevelFileName);
end;


destructor TLevel.Destroy();
begin
  inherited Destroy();
end;


function TLevel.LetterToLevelDiff(ALetter: AnsiChar): TLevelOfDifficulty;
begin
  { translate letter to level of difficulty }
  Result := LEVEL_DIF_ENUMS_ARR[ALetter];
end;


function TLevel.LetterToSnakeColoration(ALetter: AnsiChar): TSnakeColoration;
begin
  { translate letter to snake coloration }
  Result := SNAKE_COLORATION_ENUMS_ARR[ALetter];
end;


function TLevel.LetterToSnakeSpeed(ALetter: AnsiChar): TLevelOfSnakeSpeed;
begin
  { translate letter to level of snake speed }
  Result := SNAKE_SPEED_ENUMS_ARR[ALetter];
end;


function TLevel.LetterToSnakeMoveDir(ALetter: AnsiChar): TMoveDirection;
begin
  { translate letter to snake move direction }
  Result := SNAKE_MOVE_DIRECTIONS_ARR[ALetter];
end;


function TLevel.LetterToMapChar(ALetter: AnsiChar): AnsiChar;
begin
  { translate letter to level map character }
  Result := LEVEL_MAP_CHARS_ARR[ALetter];
end;


function TLevel.GetSnakeSpeed(): UInt8;
begin
  { return the converted enumeration to number }
  Result := UInt8(FSnakeInfo.Speed);
end;


function TLevel.GetMapModuleChar(AX, AY: UInt8): AnsiChar;
begin
  { return the level map module character }
  Result := FMapInfo.Map[AX, AY];
end;


function TLevel.GetGrassClump(AIndex: UInt8): TLevelGrassClumpRec;
begin
  { return the level grass clump record }
  Result := FGrassInfo.Clumps[AIndex];
end;


function TLevel.CoordsInHole(ACoords: TModuleCoords): Boolean;
var
  intToken: Integer;
begin
  { check all holes }
  for intToken := 0 to FHolesInfo.Count - 1 do
  begin
    Result :=  (ACoords.X >= FHolesInfo.Holes[intToken].Left) and (ACoords.X <= FHolesInfo.Holes[intToken].Right) and
               (ACoords.Y >= FHolesInfo.Holes[intToken].Top)  and (ACoords.Y <= FHolesInfo.Holes[intToken].Bottom);

    if Result then Exit();
  end;

  { coordinates is not in hole }
  Result := False;
end;


procedure TLevel.LoadFromFile(AFileName: AnsiString);
var
  iniLevel: TIniFile;

  procedure ReadLevelInfo();
  var
    strDifficulty: AnsiString;
  begin
    { read the name of level }
    FLevelInfo.Name := iniLevel.ReadString(LFSN_INFO, LFKN_INFO_NAME, '');
    { read the level of difficulty }
    strDifficulty := iniLevel.ReadString(LFSN_INFO, LFKN_INFO_DIFFICULTY, '');
    FLevelInfo.Difficulty := LetterToLevelDiff(strDifficulty[1]);
  end;

  procedure ReadLevelSnake();
  var
    strColoration, strDirection, strSnakeSpeed: AnsiString;
  begin
    { read the snake coloration }
    strColoration := iniLevel.ReadString(LFSN_SNAKE, LFKN_SNAKE_COLORATION, '');
    FSnakeInfo.Coloration := LetterToSnakeColoration(strColoration[1]);
    { read the snake head default coordinates }
    FSnakeInfo.DefHeadCoords.X := iniLevel.ReadInteger(LFSN_SNAKE, LFKN_SNAKE_DEF_HEAD_POS_X, 2);
    FSnakeInfo.DefHeadCoords.Y := iniLevel.ReadInteger(LFSN_SNAKE, LFKN_SNAKE_DEF_HEAD_POS_Y, 2);
    { read the snake default move direction }
    strDirection := iniLevel.ReadString(LFSN_SNAKE, LFKN_SNAKE_DEF_DIRECTION, '');
    FSnakeInfo.DefDirection := LetterToSnakeMoveDir(strDirection[1]);
    { read the level of snake speed }
    strSnakeSpeed := iniLevel.ReadString(LFSN_SNAKE, LFKN_SNAKE_SPEED, '');
    FSnakeInfo.Speed := LetterToSnakeSpeed(strSnakeSpeed[1]);
  end;

  procedure ReadLevelMap();
  var
    intX, intY: UInt8;
    strMapRow, strBoundsColor: AnsiString;
  begin
    { read the color of level bounds }
    strBoundsColor := iniLevel.ReadString(LFSN_MAP, LFKN_MAP_BOUNDS_COLOR, '');
    FMapInfo.BoundsColor := Ord(strBoundsColor[1]) - LEVEL_BOUNDS_COLOR_CHAR_DIFF;

    { read the level map }
    for intY := LBV_TOP to LBV_BOTTOM do
    begin
      strMapRow := iniLevel.ReadString(LFSN_MAP, IntToStr(intY), '');

      for intX := LBV_LEFT to LBV_RIGHT do
        FMapInfo.Map[intX, intY] := LetterToMapChar(strMapRow[intX]);
    end;
  end;

  procedure ReadLevelHoles();
  var
    strHole: AnsiString;
    intToken: Integer;
  begin
    { read the holes count }
    FHolesInfo.Count := iniLevel.ReadInteger(LFSN_HOLES, LFKN_HOLES_COUNT, 0);
    SetLength(FHolesInfo.Holes, FHolesInfo.Count);

    { read the holes }
    for intToken := 0 to FHolesInfo.Count - 1 do
    begin
      strHole := LFSN_HOLE + IntToStr(intToken + 1);
      { read the hole dimensions }
      FHolesInfo.Holes[intToken].Left   := iniLevel.ReadInteger(strHole, LFKN_HOLE_LEFT, 1);
      FHolesInfo.Holes[intToken].Right  := iniLevel.ReadInteger(strHole, LFKN_HOLE_RIGHT, 1);
      FHolesInfo.Holes[intToken].Top    := iniLevel.ReadInteger(strHole, LFKN_HOLE_TOP, 1);
      FHolesInfo.Holes[intToken].Bottom := iniLevel.ReadInteger(strHole, LFKN_HOLE_BOTTOM, 1);
    end;
  end;

  procedure ReadLevelGrassClumps();
  var
    strClump, strGrassColor: AnsiString;
    intToken: Integer;
  begin
    { read the grass clumps count }
    FGrassInfo.Count := iniLevel.ReadInteger(LFSN_GRASS_CLUMPS, LFKN_CLUMPS_COUNT, 0);
    SetLength(FGrassInfo.Clumps, FGrassInfo.Count);

    { read the grass clumps }
    for intToken := 0 to FGrassInfo.Count - 1 do
    begin
      strClump := LFSN_GRASS_CLUMP + IntToStr(intToken + 1);
      { read clump color }
      strGrassColor := iniLevel.ReadString(strClump, LFKN_CLUMP_COLOR, '');
      FGrassInfo.Clumps[intToken].Color := Ord(strGrassColor[1]) - LEVEL_GRASS_COLOR_CHAR_DIFF;
      { read clump dimensions }
      FGrassInfo.Clumps[intToken].Left   := iniLevel.ReadInteger(strClump, LFKN_CLUMP_LEFT, 1);
      FGrassInfo.Clumps[intToken].Right  := iniLevel.ReadInteger(strClump, LFKN_CLUMP_RIGHT, 1);
      FGrassInfo.Clumps[intToken].Top    := iniLevel.ReadInteger(strClump, LFKN_CLUMP_TOP, 1);
      FGrassInfo.Clumps[intToken].Bottom := iniLevel.ReadInteger(strClump, LFKN_CLUMP_BOTTOM, 1);
    end;
  end;

begin
  iniLevel := TIniFile.Create(AFileName);
  try
    { read the level general information }
    ReadLevelInfo();
    { read the level snake }
    ReadLevelSnake();
    { read the level map }
    ReadLevelMap();
    { read the level holes }
    ReadLevelHoles();
    { read the level grass clumps }
    ReadLevelGrassClumps();
  finally
    iniLevel.Free();
  end;
end;


function TLevel.CanMove(ACoords: TModuleCoords): Boolean;
begin
  { check the bounds colision }
  Result := (FMapInfo.Map[ACoords.X, ACoords.Y] = #32) and not CoordsInHole(ACoords);
end;


function TLevel.IsInLevel(ACoords: TModuleCoords): Boolean;
begin
  { check the holes }
  Result := not CoordsInHole(ACoords);
end;


{ ----- end implementation ------------------------------------------------------------------------------------ }


end.

