{

    Snake.pp                           last modified: 5 July 2014

    Copyright (C) Jaroslaw Baran, furious programming 2011 - 2014.
    All rights reserved.
   __________________________________________________________________________

    This unit is a part of the SnakeASCII console game.
   __________________________________________________________________________

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
   __________________________________________________________________________

}

unit Snake;

{$MODE OBJFPC}{$H+}{$PACKENUM 1}

interface

uses
  Converter, Common;


type
  { snake open gates }
  TSnakeGates = object
  private
    FGatesArr: array of TModuleCoords;
    FCount: UInt8;
  public
    procedure Add(ACoords: TModuleCoords);
    procedure Remove(ACoords: TModuleCoords);
  public
    function IsOpenGate(ACoords: TModuleCoords): Boolean;
  public
    procedure Reset();
  end;

type
  { snake class }
  TSnake = class(TObject)
  private
    FModules: TSnakeModulesArr;
    FGates: TSnakeGates;
    FLength: TSnakeLength;
    FCurrDirection: TMoveDirection;
    FPermTurbo: Boolean;
  private
    procedure BuildSnake(ALength: TSnakeLength; ACoords: TModuleCoords; ADirection: TMoveDirection);
  private
    function GetModule(AIndex: TSnakeLength): TSnakeModuleRec;
    function GetModuleType(ACoords: TModuleCoords): TSnakeModuleBasicType;
    function GetHeadCoords(): TModuleCoords;
    function IsHeadOnOtherModule(): Boolean;
  public
    constructor Create(ACoords: TModuleCoords; ADirection: TMoveDirection);
    destructor Destroy(); override;
  public
    function IsGate(ACoords: TModuleCoords): Boolean;
    function IsNotOnSnake(ACoords: TModuleCoords): Boolean;
    function CanMove(ACoords: TModuleCoords): Boolean;
  public
    procedure Move(ADirection: TMoveDirection; ACaught: Boolean);
    procedure Reset(ACoords: TModuleCoords; ADirection: TMoveDirection);
    procedure InvertPermanentTurbo();
  public
    property Length: TSnakeLength read FLength;
    property Direction: TMoveDirection read FCurrDirection;
    property HeadCoords: TModuleCoords read GetHeadCoords;
    property PermanentTurbo: Boolean read FPermTurbo;
  public
    property Module[AIndex: TSnakeLength]: TSnakeModuleRec read GetModule; default;
    property ModuleType[ACoords: TModuleCoords]: TSnakeModuleBasicType read GetModuleType;
  end;


{ ----- end interface ----------------------------------------------------------------------------------------- }


implementation


{ ----- TSnakeGates object ------------------------------------------------------------------------------------ }


procedure TSnakeGates.Add(ACoords: TModuleCoords);
begin
  { add the new open gate to gates array }
  SetLength(FGatesArr, FCount + 1);
  FGatesArr[FCount] := ACoords;
  { increment gates counter }
  Inc(FCount);
end;


procedure TSnakeGates.Remove(ACoords: TModuleCoords);
var
  intToken: UInt8;
begin
  { find deleting gate }
  for intToken := 0 to FCount - 1 do
    { compare coordinates }
    if TConverter.SameModulesCoords(ACoords, FGatesArr[intToken]) then
    begin
      { decrement gates counter }
      Dec(FCount);

      { move gates memory if is not the last gate in array }
      if intToken < FCount then
        Move(FGatesArr[intToken + 1], FGatesArr[intToken], (FCount - intToken) * SizeOf(TModuleCoords));

      { decrease the gates array size }
      SetLength(FGatesArr, FCount);
      Exit;
    end;
end;


function TSnakeGates.IsOpenGate(ACoords: TModuleCoords): Boolean;
var
  mcToken: TModuleCoords;
begin
  { find the gate }
  for mcToken in FGatesArr do
    { compare coordinates }
    if TConverter.SameModulesCoords(ACoords, mcToken) then
      { same modules coordinates }
      Exit(True);

  { no gate found }
  Result := False;
end;


procedure TSnakeGates.Reset();
begin
  { reinitialize gates array }
  FCount := 0;
  SetLength(FGatesArr, FCount);
end;


{ ----- TSnake class ------------------------------------------------------------------------------------------ }


constructor TSnake.Create(ACoords: TModuleCoords; ADirection: TMoveDirection);
begin
  inherited Create();

  { build snake }
  BuildSnake(SNAKE_INITIAL_LENGTH, ACoords, ADirection);
end;


destructor TSnake.Destroy();
begin
  inherited Destroy();
end;


procedure TSnake.BuildSnake(ALength: TSnakeLength; ACoords: TModuleCoords; ADirection: TMoveDirection);
const
  { snake modules types array }
  SNAKE_MODULE_TYPES: array [TMoveDirection] of TSnakeModuleBasicType = (
    smbtVertical, smbtVertical, smbtHorzizontal, smbtHorzizontal, smbtNone
  );
var
  slToken:  TSnakeLength  absolute ALength;
  smcToken: TModuleCoords absolute ACoords;
begin
  { set the snake fields }
  FLength := ALength;
  FCurrDirection := ADirection;
  FPermTurbo := False;
  { set the token value to head module index }
  Dec(slToken);

  { fill snake head module }
  FModules[slToken].ModuleType := smbtHead;
  FModules[slToken].Coords     := smcToken;
  FModules[slToken].GoldModule := False;

  { fill snake tail modules }
  repeat
    Dec(slToken);

    { update snake module coordinates }
    case ADirection of
      mdUp:    Inc(smcToken.Y);
      mdDown:  Dec(smcToken.Y);
      mdLeft:  Inc(smcToken.X);
      mdRight: Dec(smcToken.X);
    end;

    { fill snake module }
    FModules[slToken].ModuleType := SNAKE_MODULE_TYPES[ADirection];
    FModules[slToken].Coords     := smcToken;
    FModules[slToken].GoldModule := False;
  until slToken = 0;

  { initialize gates array }
  FGates.Reset();
end;


function TSnake.GetModule(AIndex: TSnakeLength): TSnakeModuleRec;
begin
  { return module from modules array }
  Result := FModules[AIndex];
end;


function TSnake.GetModuleType(ACoords: TModuleCoords): TSnakeModuleBasicType;
var
  slToken: TSnakeLength;
begin
  { find module in coordinates }
  for slToken := FLength - 2 downto 1 do
    { compare coordinates }
    if TConverter.SameModulesCoords(ACoords, FModules[slToken].Coords) then
      Exit(FModules[slToken].ModuleType);

  { no module found (just in case) }
  Result := smbtNone;
end;


function TSnake.GetHeadCoords(): TModuleCoords;
begin
  { return the head of snake coordinates }
  Result := FModules[FLength - 1].Coords;
end;


function TSnake.IsHeadOnOtherModule(): Boolean;
var
  slToken: TSnakeLength;
begin
  { find module in coordinates without head }
  for slToken := FLength - 2 downto 1 do
    { compare coordinates }
    if TConverter.SameModulesCoords(FModules[slToken].Coords, FModules[FLength - 1].Coords) then
      Exit(True);

  { no module found }
  Result := False;
end;


function TSnake.IsGate(ACoords: TModuleCoords): Boolean;
begin
  { check the gates }
  Result := FGates.IsOpenGate(ACoords);
end;


function TSnake.IsNotOnSnake(ACoords: TModuleCoords): Boolean;
var
  slToken: TSnakeLength;
begin
  { check the snake modules coordinates }
  for slToken := 0 to FLength - 1 do
    if TConverter.SameModulesCoords(ACoords, FModules[slToken].Coords) then
      Exit(False);

  { is not on the snake }
  Result := True;
end;


function TSnake.CanMove(ACoords: TModuleCoords): Boolean;
var
  slToken: TSnakeLength;
begin
  { check all snake body modules (without head module) }
  for slToken := 0 to FLength - 3 do
    { compare coordinates }
    if TConverter.SameModulesCoords(ACoords, FModules[slToken].Coords) then
      { gold module and not a corner state }
      Exit(FModules[slToken].GoldModule and not TConverter.CornerModule(FModules[slToken].ModuleType));

  { is free module }
  Result := True;
end;


procedure TSnake.Move(ADirection: TMoveDirection; ACaught: Boolean);
var
  slToken: TSnakeLength = 0;
  mdHeadToNeck, mdNeckToNext: TMoveDirection;
begin
  { update the current move direction }
  FCurrDirection := ADirection;

  { if snake does not caught the gold }
  if not ACaught then
    { tail is on the gate module }
    if FGates.IsOpenGate(FModules[0].Coords) then
      { close the gate in tail module }
      FGates.Remove(FModules[0].Coords);

  { caught the gold state }
  if ACaught then
  begin
    { increase snake length }
    Inc(FLength);
    { set the modules token value }
    slToken := FLength - 1;
    { set the new head data }
    FModules[slToken].ModuleType := smbtHead;
    FModules[slToken].Coords     := FModules[slToken - 1].Coords;
    FModules[slToken].GoldModule := True;
  end
  else
  begin
    { rewrite body modules coordinates }
    while slToken < FLength - 1 do
    begin
      FModules[slToken].Coords     := FModules[slToken + 1].Coords;
      FModules[slToken].GoldModule := FModules[slToken + 1].GoldModule;
      Inc(slToken);
    end;

    { set the new head gold module state }
    FModules[slToken].GoldModule := False;
  end;

  { update the new head module coordinates }
  case ADirection of
    mdUp:    Dec(FModules[slToken].Coords.Y);
    mdDown:  Inc(FModules[slToken].Coords.Y);
    mdLeft:  Dec(FModules[slToken].Coords.X);
    mdRight: Inc(FModules[slToken].Coords.X);
  end;

  { head is on the other module }
  if IsHeadOnOtherModule() then
    { open the gate }
    FGates.Add(FModules[slToken].Coords);

  { caught the gold state }
  if ACaught then
    slToken := FLength - 2
  else
  begin
    slToken := 0;

    { rewrite body modules types }
    while slToken < FLength - 2 do
    begin
      FModules[slToken].ModuleType := FModules[slToken + 1].ModuleType;
      Inc(slToken);
    end;
  end;

  { get the head, neck and next modules coordinates relationship }
  mdHeadToNeck := TConverter.CoordsRelationshipToMoveDirection(FModules[slToken + 1].Coords, FModules[slToken].Coords);
  mdNeckToNext := TConverter.CoordsRelationshipToMoveDirection(FModules[slToken].Coords, FModules[slToken - 1].Coords);

  { update the neck module type }
  FModules[slToken].ModuleType := TConverter.MoveDirectionsRelationshipToModuleBasicType(mdHeadToNeck, mdNeckToNext);
end;


procedure TSnake.Reset(ACoords: TModuleCoords; ADirection: TMoveDirection);
begin
  { rebuild snake modules }
  BuildSnake(SNAKE_INITIAL_LENGTH, ACoords, ADirection);
end;


procedure TSnake.InvertPermanentTurbo();
begin
  { invert the PermanentTurbo state }
  FPermTurbo := not FPermTurbo;
end;


{ ----- end implementation ------------------------------------------------------------------------------------ }


end.

